import json
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

import humanfriendly
import pydantic

from ...components.container.models import ContainerConfig
from ...utils import DockerCamelModel, all_fields_optional


@all_fields_optional
class ImageHealthcheck(DockerCamelModel):
    test: List[str]
    interval: int
    timeout: int
    retries: int
    start_period: int


@all_fields_optional
class ImageGraphDriver(DockerCamelModel):
    name: str
    data: Any


@all_fields_optional
class ImageRootFS(DockerCamelModel):
    type: str
    layers: List[str]
    base_layer: str


@all_fields_optional
class ImageInspectResult(DockerCamelModel):
    id: str
    repo_tags: List[str]
    repo_digests: List[str]
    parent: str
    comment: str
    created: datetime
    container: str
    container_config: ContainerConfig
    docker_version: str
    author: str
    config: ContainerConfig
    architecture: str
    os: str
    os_version: str
    size: int
    virtual_size: int
    graph_driver: ImageGraphDriver
    root_fs: ImageRootFS
    metadata: Dict[str, str]


@all_fields_optional
class ImageHistoryLayer(DockerCamelModel):
    id: str
    created_at: datetime
    created_by: str
    created_since: str
    comment: str
    size: float

    @staticmethod
    def parse_json(json_str: str) -> 'ImageHistoryLayer':
        d: dict = json.loads(json_str)
        return ImageHistoryLayer(
            id=d["ID"],
            created_at=d["CreatedAt"],
            created_by=d["CreatedBy"],
            created_since=d["CreatedSince"],
            comment=d["Comment"],
            size=humanfriendly.parse_size(d["Size"]),
        )


class LayerPullStatus(Enum):
    UNKNOWN = "unknown"
    PULLING = "pulling fs layer"
    WAITING = "waiting"
    DOWNLOADING = "downloading"
    DOWNLOADED = "download complete"
    VERIFYING = "verifying checksum"
    EXTRACTING = "extracting"
    PULLED = "pull complete"
    EXISTS = "already exists"

    @classmethod
    def from_string(cls, s: str) -> 'LayerPullStatus':
        try:
            return LayerPullStatus(s.lower().strip())
        except Exception:
            return LayerPullStatus.UNKNOWN


class LayerPullProgress(pydantic.BaseModel):
    layer_id: str
    status: LayerPullStatus
    progress: Optional[float] = None



from pathlib import Path
from typing import List, Optional, Union

from ....client_config import DockerCLICaller
from ....exceptions import DockerException
from ....utils import run
from .models import Manifest


class ImagetoolsCLI(DockerCLICaller):
    def inspect(self, name: str) -> Optional[Manifest]:
        """Returns the manifest of a Docker image in a registry without pulling it"""
        full_cmd = self.docker_cmd + ["buildx", "imagetools", "inspect", "--raw", name]
        try:
            result = run(full_cmd)
        except DockerException as e:
            # manifest not found
            if "ERROR" in e.stderr and "not found" in e.stderr:
                return None
            raise e
        return Manifest.parse_raw(result)

    def create(
        self,
        sources: List[str] = [],
        tags: List[str] = [],
        append: bool = False,
        files: List[Union[str, Path]] = [],
        dry_run: bool = False,
        builder: Optional[str] = None,
    ) -> Optional[Manifest]:
        """
        Create a new manifest list based on source manifests.
        The source manifests can be manifest lists or single platform distribution manifests and
        must already exist in the registry where the new manifest is created.
        If only one source is specified, create performs a carbon copy.

        The CLI docs is [here](https://docs.docker.com/engine/reference/commandline/buildx_imagetools_create/)
        and it contains a lot more information.

        # Arguments
            sources: The sources manifest to create, change
            append: Append to existing manifest
            dry_run: Show final image instead of pushing
            files: Read source descriptor from file
            builder: The builder to use.
        """
        if not isinstance(sources, list):
            raise TypeError(
                "The argument 'sources' of the function docker.buildx.imagetools.create() must be a list of strings."
            )
        if not isinstance(tags, list):
            raise TypeError(
                "The argument 'tags' of the function docker.buildx.imagetools.create() must be a list of strings."
            )
        if not isinstance(files, list):
            raise TypeError(
                "The argument 'files' of the function docker.buildx.imagetools.create() must be a list of strings."
            )

        full_cmd = self.docker_cmd + ["buildx", "imagetools", "create"]
        for tag in tags:
            full_cmd.add_simple_arg("--tag", tag)
        for file in files:
            full_cmd.add_simple_arg("--file", file)
        full_cmd.add_simple_arg("--builder", builder)
        full_cmd.add_flag("--append", append)
        full_cmd.add_flag("--dry-run", dry_run)
        full_cmd += sources

        result = run(full_cmd)
        if dry_run:
            return Manifest.parse_raw(result)



from typing import List, Optional


class DockerException(Exception):
    def __init__(
        self,
        command_launched: List[str],
        return_code: int,
        stdout: Optional[bytes] = None,
        stderr: Optional[bytes] = None,
    ):
        self.docker_command: List[str] = command_launched
        self.return_code: int = return_code
        if stdout is None:
            self.stdout: Optional[str] = None
        else:
            self.stdout: Optional[str] = stdout.decode()
        if stderr is None:
            self.stderr: Optional[str] = None
        else:
            self.stderr: Optional[str] = stderr.decode()
        command_launched_str = " ".join(command_launched)
        error_msg = (
            f"The docker command executed was `{command_launched_str}`.\n"
            f"It returned with code {return_code}\n"
        )
        if stdout is not None:
            error_msg += f"The content of stdout is '{self.stdout}'\n"
        else:
            error_msg += (
                "The content of stdout can be found above the "
                "stacktrace (it wasn't captured).\n"
            )
        if stderr is not None:
            error_msg += f"The content of stderr is '{self.stderr}'\n"
        else:
            error_msg += (
                "The content of stderr can be found above the "
                "stacktrace (it wasn't captured)."
            )
        super().__init__(error_msg)


class NoSuchContainer(DockerException):
    pass


class NoSuchImage(DockerException):
    pass


class NoSuchService(DockerException):
    pass


class NotASwarmManager(DockerException):
    pass


class NoSuchVolume(DockerException):
    pass


class NoSuchManifest(DockerException):
    pass


class TemporaryFailureInNameResolution(DockerException):
    pass



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from ...client_config import DockerCLICaller
from ...components.service import cli_wrapper as service_cli_wrapper
from ...components.task import cli_wrapper as task_cli_wrapper
from ...utils import ValidPath, read_env_files, run, to_list


class Stack:
    def __init__(self, client_config, name):
        self.client_config = client_config
        self.name = name

    def __str__(self):
        return self.name

    def __eq__(self, other: Any):
        if isinstance(other, Stack):
            return self.client_config == other.client_config and self.name == other.name
        else:
            return False

    def __repr__(self):
        return f"dockertown.Stack(name='{self.name}')"

    def remove(self) -> None:
        StackCLI(self.client_config).remove(self)

    def ps(self) -> List[task_cli_wrapper.Task]:
        return StackCLI(self.client_config).ps(self)

    def services(self) -> List[service_cli_wrapper.Service]:
        return StackCLI(self.client_config).services(self)


ValidStack = Union[str, Stack]


class StackCLI(DockerCLICaller):
    def deploy(
        self,
        name: str,
        compose_files: Union[ValidPath, List[ValidPath]] = [],
        orchestrator: Optional[str] = None,
        prune: bool = False,
        resolve_image: str = "always",
        with_registry_auth: bool = False,
        env_files: List[ValidPath] = [],
        variables: Dict[str, str] = {},
    ) -> Stack:
        """Deploys a stack.

        # Arguments
            name: The name of the stack to deploy. Mandatory.
            compose_files: One or more docker-compose files. If there are more than
            one, they will be fused together.
            orchestrator: The orchestrator to use, `"swarm" or "kubernetes" or "all".
            prune: Prune services that are no longer referenced
            resolve_image: Query the registry to resolve image digest
                and supported platforms `"always"|"changed"|"never"` (default `"always"`).
                Note that if the registry cannot be queried when using `"always"`, it's
                going to try to use images present locally on the nodes.
            with_registry_auth: Send registry authentication details to Swarm agents.
                Required if you need to run `docker login` to pull the docker images
                in your stack.
            env_files: Similar to `.env` files in docker-compose. Loads `variables` from
                `.env` files. If both `env_files` and `variables` are used, `variables`
                have priority. This behavior is similar to the one you would experience with
                compose.
            variables: A dict dictating by what to replace the variables declared in
                the docker-compose files. In the docker CLI, you would use
                environment variables for this.

        # Returns
            A `dockertown.Stack` object.
        """
        full_cmd = self.docker_cmd + ["stack", "deploy"]

        full_cmd.add_args_list("--compose-file", compose_files)
        full_cmd.add_simple_arg("--orchestrator", orchestrator)
        full_cmd.add_flag("--prune", prune)
        full_cmd.add_simple_arg("--resolve-image", resolve_image)
        full_cmd.add_flag("--with-registry-auth", with_registry_auth)
        full_cmd.append(name)

        env = read_env_files([Path(x) for x in env_files])
        env.update(variables)

        run(full_cmd, capture_stdout=False, env=env)
        return Stack(self.client_config, name)

    def list(self) -> List[Stack]:
        """Returns a list of `dockertown.Stack`

        # Returns
            A `List[dockertown.Stack]`.
        """
        full_cmd = self.docker_cmd + ["stack", "ls", "--format", "{{.Name}}"]
        stacks_names = run(full_cmd).splitlines()
        return [Stack(self.client_config, name) for name in stacks_names]

    def ps(self, x: ValidStack) -> List[task_cli_wrapper.Task]:
        """Returns the list of swarm tasks in this stack.

        ```python
        from dockertown import docker

        tasks = docker.stack.ps("my-stack")
        print(tasks[0].desired_state)
        # running
        ```

        # Arguments
            x: A stack . It can be name or a `dockertown.Stack` object.

        # Returns
            `List[dockertown.Task]`
        """
        full_cmd = self.docker_cmd + ["stack", "ps", "--quiet", "--no-trunc", x]

        ids = run(full_cmd).splitlines()
        return [
            task_cli_wrapper.Task(self.client_config, id_, is_immutable_id=True)
            for id_ in ids
        ]

    def remove(self, x: Union[ValidStack, List[ValidStack]]) -> None:
        """Removes one or more stacks.

        # Arguments
            x: One or more stacks, empty list means nothing will be done.

        """
        if x == []:
            return
        full_cmd = self.docker_cmd + ["stack", "remove"] + to_list(x)
        run(full_cmd)

    def services(self, stack: ValidStack) -> List[service_cli_wrapper.Service]:
        """List the services present in the stack.

        # Arguments
            stack: A docker stack or the name of a stack.

        # Returns
            A `List[dockertown.Stack]`
        """
        full_cmd = self.docker_cmd + ["stack", "services", "--quiet", stack]
        ids = run(full_cmd).splitlines()
        return [service_cli_wrapper.Service(self.client_config, id_) for id_ in ids]



from typing import Any, Dict, List, Optional, Union

from ...client_config import ClientConfig, DockerCLICaller, ReloadableObjectFromJson
from ...utils import ValidPath, format_dict_for_cli, run, to_list
from .models import SecretInspectResult


class Secret(ReloadableObjectFromJson):
    def __init__(
        self, client_config: ClientConfig, reference: str, is_immutable_id=False
    ):
        super().__init__(client_config, "id", reference, is_immutable_id)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove()

    def _fetch_inspect_result_json(self, reference):
        return run(self.docker_cmd + ["secret", "inspect", reference])

    def _parse_json_object(self, json_object: Dict[str, Any]) -> SecretInspectResult:
        return SecretInspectResult.parse_obj(json_object)

    def _get_inspect_result(self) -> SecretInspectResult:
        """Only there to allow tools to know the return type"""
        return super()._get_inspect_result()

    @property
    def id(self) -> str:
        return self._get_immutable_id()

    def remove(self):
        """Remove this Docker secret.

        See the [`docker.secret.remove`](../sub-commands/secret.md#remove) command for
        information about the arguments.
        """
        SecretCLI(self.client_config).remove(self)


ValidSecret = Union[Secret, str]


class SecretCLI(DockerCLICaller):
    def create(
        self,
        name: str,
        file: ValidPath,
        driver: Optional[str] = None,
        labels: Dict[str, str] = {},
        template_driver: Optional[str] = None,
    ) -> Secret:
        """Creates a `dockertown.Secret`.

        # Returns
            A `dockertown.Secret` object.
        """
        full_cmd = self.docker_cmd + ["secret", "create"]
        full_cmd.add_simple_arg("--driver", driver)
        full_cmd.add_args_list("--label", format_dict_for_cli(labels))
        full_cmd.add_simple_arg("--template-driver", template_driver)
        full_cmd += [name, file]
        return Secret(self.client_config, run(full_cmd), is_immutable_id=True)

    def inspect(self, x: Union[str, List[str]]) -> Union[Secret, List[Secret]]:
        """Returns one or more `dockertown.Secret` based on an ID or name.

        # Arguments
            x: One or more IDs/names.
        """
        if isinstance(x, list):
            return [Secret(self.client_config, reference) for reference in x]
        else:
            return Secret(self.client_config, x)

    def list(self, filters: Dict[str, str] = {}) -> List[Secret]:
        """Returns all secrets as a `List[dockertown.Secret]`."""
        full_cmd = self.docker_cmd + ["secret", "list", "--quiet"]
        full_cmd.add_args_list("--filter", format_dict_for_cli(filters))
        ids = run(full_cmd).splitlines()
        return [Secret(self.client_config, id_, is_immutable_id=True) for id_ in ids]

    def remove(self, x: Union[ValidSecret, List[ValidSecret]]) -> None:
        """Removes one or more secrets

        # Arguments
            x: One or more secrets.
                Name, ids or `dockertown.Secret` objects are valid inputs.
        """
        if x == []:
            return
        full_cmd = self.docker_cmd + ["secret", "remove"] + to_list(x)
        run(full_cmd)



from datetime import datetime
from typing import Any, Dict, List

import pydantic

from ...utils import DockerCamelModel, all_fields_optional


@all_fields_optional
class ObjectVersion(DockerCamelModel):
    index: int


@all_fields_optional
class NamedResourceSpec(DockerCamelModel):
    kind: str
    value: str


@all_fields_optional
class DiscreteResourceSpec(DockerCamelModel):
    kind: str
    value: int


@all_fields_optional
class AssignedGenericResources(DockerCamelModel):
    named_resource_spec: NamedResourceSpec
    discrete_resource_spec: DiscreteResourceSpec


@all_fields_optional
class ContainerStatus(DockerCamelModel):
    container_id: str = pydantic.Field(alias="ContainerID")
    pid: int = pydantic.Field(alias="PID")
    exit_code: int


@all_fields_optional
class TaskStatus(DockerCamelModel):
    timestamp: datetime
    state: str
    message: str
    err: str
    container_status: ContainerStatus


@all_fields_optional
class LogDriver(DockerCamelModel):
    name: str
    options: Dict[str, str]


@all_fields_optional
class NetworkAttachmentConfig(DockerCamelModel):
    target: str
    aliases: List[str]
    driver_opts: Dict[str, str]


@all_fields_optional
class Platform(DockerCamelModel):
    architecture: str
    os: str = pydantic.Field(alias="OS")


@all_fields_optional
class Spread(DockerCamelModel):
    spread_descriptor: str


@all_fields_optional
class Placement(DockerCamelModel):
    constraints: List[str]
    preferences: List[Spread]
    max_replicas: int
    platforms: List[Platform]


@all_fields_optional
class RestartPolicy(DockerCamelModel):
    condition: str
    delay: int
    max_attempts: int
    window: int


@all_fields_optional
class PluginPrivilege(DockerCamelModel):
    name: str
    description: str
    value: List[str]


@all_fields_optional
class PluginSpec(DockerCamelModel):
    name: str
    remote: str
    disabled: bool
    plugin_privilege: List[PluginPrivilege]


@all_fields_optional
class ContainerSpec(DockerCamelModel):
    image: str
    labels: Dict[str, str]
    command: List[str]
    args: List[str]
    hostname: str
    env: List[str]
    dir: str
    user: str
    groups: List[str]
    privileges: Any
    tty: bool = pydantic.Field(alias="TTY")
    open_stdin: bool
    read_only: bool
    mounts: List[Any]
    stop_signal: str
    stop_grace_period: int
    health_check: Any
    hosts: List[str]
    dns_config: Any
    secrets: List[Any]
    configs: List[Any]
    isolation: str
    init: bool
    sysctls: Any


@all_fields_optional
class NetworkAttachmentSpec(DockerCamelModel):
    container_id: str = pydantic.Field(alias="ContainerID")


@all_fields_optional
class ResourceObject(DockerCamelModel):
    nano_cpus: int = pydantic.Field(alias="NanoCPUs")
    memory_bytes: int
    generic_resources: List[AssignedGenericResources]


@all_fields_optional
class Resources(DockerCamelModel):
    limits: ResourceObject
    reservation: ResourceObject


@all_fields_optional
class TaskSpec(DockerCamelModel):
    # TODO: set types for Any
    plugin_spec: PluginSpec
    container_spec: ContainerSpec
    network_attachment_spec: NetworkAttachmentSpec
    resources: Resources
    restart_policy: Any
    placement: Placement
    force_update: int
    runtime: str
    networks: List[NetworkAttachmentConfig]
    log_driver: LogDriver


@all_fields_optional
class TaskInspectResult(DockerCamelModel):
    id: str = pydantic.Field(alias="ID")
    version: ObjectVersion
    created_at: datetime
    updated_at: datetime
    name: str
    labels: Dict[str, str]
    spec: TaskSpec
    service_id: str = pydantic.Field(alias="ServiceID")
    slot: int
    node_id: str = pydantic.Field(alias="NodeID")
    assigned_generic_resources: List[AssignedGenericResources]
    status: TaskStatus
    desired_state: str



from typing import Optional

import typer

from . import docker
from .download_binaries import download_docker_cli

app = typer.Typer()


volume_app = typer.Typer()


@volume_app.command()
def copy(source: str, destination: str):
    if ":" in source:
        source_volume, source_path = source.split(":")
        docker.volume.copy((source_volume, source_path), destination)
    elif ":" in destination:
        destination_volume, destination_path = destination.split(":")
        docker.volume.copy(source, (destination_volume, destination_path))
    else:
        raise ValueError("No volume was specified. The format is 'volume:path'")


image_app = typer.Typer()


@image_app.command()
def copy_from(docker_image: str, source: str, destination: str, pull: str = "missing"):
    docker.image.copy_from(docker_image, source, destination, pull)


@image_app.command()
def copy_to(
    docker_image: str,
    source: str,
    destination: str,
    new_tag: Optional[str] = None,
    push: bool = False,
    pull: str = "missing",
):
    docker.image.copy_to(docker_image, source, destination, new_tag, pull)
    if push:
        docker.image.push(new_tag)


@app.command()
def download_cli():
    download_docker_cli()


@app.command()
def download_buildx():
    raise NotImplementedError("Downloading the buildx binary isn't supported yet.")


app.add_typer(volume_app, name="volume")
app.add_typer(image_app, name="image")


def main():
    app()



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, overload

from ...client_config import ClientConfig, DockerCLICaller, ReloadableObjectFromJson
from ...utils import run, to_list
from .models import ContextEndpoint, ContextInspectResult, ContextStorage


class Context(ReloadableObjectFromJson):
    def __init__(
        self,
        client_config: ClientConfig,
        reference: Optional[str],
        is_immutable_id=False,
    ):
        super().__init__(client_config, "name", reference, is_immutable_id)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove(force=True)

    def _fetch_inspect_result_json(self, reference: Optional[str]):
        full_cmd = self.docker_cmd + ["context", "inspect"]
        if reference is not None:
            full_cmd.append(reference)
        return run(full_cmd)

    def _parse_json_object(self, json_object: Dict[str, Any]):
        return ContextInspectResult.parse_obj(json_object)

    def _get_inspect_result(self) -> ContextInspectResult:
        """Only there to allow tools to know the return type"""
        return super()._get_inspect_result()

    @property
    def name(self) -> str:
        return self._get_immutable_id()

    @property
    def metadata(self) -> Dict[str, str]:
        return self._get_inspect_result().metadata

    @property
    def endpoints(self) -> Dict[str, ContextEndpoint]:
        return self._get_inspect_result().endpoints

    @property
    def tls_material(self) -> dict:
        return self._get_inspect_result().tls_material

    @property
    def storage(self) -> ContextStorage:
        return self._get_inspect_result().storage

    def remove(self, force: bool = False) -> None:
        """Removes this context"""
        ContextCLI(self.client_config).remove(self, force)

    def update(self) -> None:
        """Update this context, not yet implemented."""
        raise NotImplementedError
        ContextCLI(self.client_config).update(self)

    def use(self) -> None:
        """Use this context"""
        ContextCLI(self.client_config).use(self)

    def __repr__(self):
        return f"dockertown.Context(name='{self.name}', endpoints={self.endpoints})"


ValidContext = Union[Context, str]


@dataclass
class DockerContextConfig:
    from_: Optional[ValidContext] = None
    host: Optional[str] = None
    certificate_authority: Union[Path, str, None] = None
    certificate: Union[Path, str, None] = None
    key: Union[Path, str, None] = None
    skip_tls_verify: bool = False

    def format_for_docker_cli(self) -> str:
        list_of_args = []
        if self.from_ is not None:
            list_of_args.append(f"from={self.from_}")
        if self.host is not None:
            list_of_args.append(f"host={self.host}")
        if self.certificate_authority is not None:
            list_of_args.append(f"ca={self.certificate_authority}")
        if self.certificate is not None:
            list_of_args.append(f"cert={self.certificate}")
        if self.key is not None:
            list_of_args.append(f"key={self.key}")
        list_of_args.append(f"skip-tls-verify={self.skip_tls_verify}")

        return ",".join(list_of_args)


@dataclass
class KubernetesContextConfig:
    from_: Optional[ValidContext] = None
    config_file: Union[str, Path, None] = None
    context_override: Optional[str] = None
    namespace_override: Optional[str] = None

    def format_for_docker_cli(self) -> str:
        list_of_args = []
        if self.from_ is not None:
            list_of_args.append(f"from={self.from_}")
        if self.config_file is not None:
            list_of_args.append(f"config-file={self.config_file}")
        if self.context_override is not None:
            list_of_args.append(f"context-override={self.context_override}")
        if self.namespace_override is not None:
            list_of_args.append(f"namespace-override={self.namespace_override}")
        return ",".join(list_of_args)


class ContextCLI(DockerCLICaller):
    def create(
        self,
        context_name: str,
        default_stack_orchestrator: Optional[str] = None,
        description: Optional[str] = None,
        from_: Optional[ValidContext] = None,
        docker: Union[Dict[str, Any], DockerContextConfig, None] = None,
        kubernetes: Union[Dict[str, Any], KubernetesContextConfig, None] = None,
    ) -> Context:
        """Creates a new context

        # Arguments
            context: name of the context to create
            default_stack_orchestrator: Default orchestrator for stack operations to use with this context (swarm|kubernetes|all)
            description: Description of the context
            docker: Set the docker endpoint, you can use a dict of a class to
                specify the options. The class is `dockertown.DockerContextConfig`.
            from_: Create context from a named context
            kubernetes: Set the kubernetes endpoint. You can use a dict or a class to specify the options. The class
                is `dockertown.KubernetesContextConfig`.
        """
        if isinstance(docker, dict):
            docker = DockerContextConfig(**docker)
        if isinstance(kubernetes, dict):
            kubernetes = KubernetesContextConfig(**kubernetes)

        full_cmd = self.docker_cmd + ["context", "create"]

        full_cmd.add_simple_arg(
            "--default-stack-orchestrator", default_stack_orchestrator
        )
        full_cmd.add_simple_arg("--description", description)
        full_cmd.add_simple_arg("--from", from_)
        if docker is not None:
            full_cmd.add_simple_arg("--docker", docker.format_for_docker_cli())
        if kubernetes is not None:
            full_cmd.add_simple_arg("--kubernetes", kubernetes.format_for_docker_cli())
        full_cmd.append(context_name)
        run(full_cmd)
        return self.inspect(context_name)

    @overload
    def inspect(self, x: Union[None, str]) -> Context:
        ...

    @overload
    def inspect(self, x: List[str]) -> List[Context]:
        ...

    def inspect(
        self, x: Union[None, str, List[str]] = None
    ) -> Union[Context, List[Context]]:
        """Returns the context object. If no argument is provided, returns the
        current context."""
        if isinstance(x, str) or x is None:
            return Context(self.client_config, x)
        else:
            return [Context(self.client_config, ref) for ref in x]

    def list(self) -> List[Context]:
        """List all Docker contexts available

        # Returns
            `List[dockertown.Context]`
        """

        full_cmd = self.docker_cmd + ["context", "list", "--quiet"]

        ids = run(full_cmd).splitlines()
        return [Context(self.client_config, id_, is_immutable_id=True) for id_ in ids]

    def remove(self, x: Union[ValidContext, List[ValidContext]], force: bool = False):
        """Removes one or more contexts

        # Arguments
            x: One or more contexts, empty list means no-op.
            force: Force the removal of this context
        """
        full_cmd = self.docker_cmd + ["context", "remove"]
        full_cmd.add_flag("--force", force)
        if x == []:
            return
        full_cmd += to_list(x)
        run(full_cmd)

    def update(self):
        """Not yet implemented"""
        raise NotImplementedError

    def use(self, context: ValidContext):
        """Set the default context

        # Arguments
            context: The context to set as default
        """
        full_cmd = self.docker_cmd + ["context", "use", context]
        run(full_cmd)



from ...client_config import DockerCLICaller


class TrustCLI(DockerCLICaller):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.key = None
        self.signer = None

    def inspect(self):
        """Not yet implemented"""
        raise NotImplementedError

    def revoke(self):
        """Not yet implemented"""
        raise NotImplementedError

    def sign(self):
        """Not yet implemented"""
        raise NotImplementedError



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import Field

from ...utils import DockerCamelModel, all_fields_optional


@all_fields_optional
class CPUMemoryQuotas(DockerCamelModel):
    nano_cpus: int = Field(alias="NanoCPUs")
    memory_bytes: int


@all_fields_optional
class Resources(DockerCamelModel):
    limits: CPUMemoryQuotas
    reservations: CPUMemoryQuotas


@all_fields_optional
class ContainerSpec(DockerCamelModel):
    image: str
    labels: Dict[str, str]
    privileges: Dict[str, Optional[str]]
    stop_grace_period: int
    isolation: str
    env: List[str]


@all_fields_optional
class TaskTemplate(DockerCamelModel):
    container_spec: ContainerSpec
    resources: Resources


@all_fields_optional
class ChangeConfig(DockerCamelModel):
    parallelism: int
    failure_action: str
    monitor: int
    max_failure_ratio: int
    order: str


@all_fields_optional
class ServiceSpec(DockerCamelModel):
    name: str
    labels: Dict[str, str]
    mode: Dict[str, Any]
    update_config: ChangeConfig
    rollback_config: ChangeConfig
    task_template: TaskTemplate


@all_fields_optional
class ServiceVersion(DockerCamelModel):
    index: int


@all_fields_optional
class EndpointPortConfig(DockerCamelModel):
    name: str
    protocol: str
    target_port: int
    published_port: int
    publish_mode: str


@all_fields_optional
class ServiceEndpointSpec(DockerCamelModel):
    mode: str
    ports: List[EndpointPortConfig]


@all_fields_optional
class VirtualIP(DockerCamelModel):
    network_id: str
    addr: str


@all_fields_optional
class ServiceEndpoint(DockerCamelModel):
    spec: ServiceEndpointSpec
    ports: List[EndpointPortConfig]
    virtual_ips: List[VirtualIP]


@all_fields_optional
class ServiceUpdateStatus(DockerCamelModel):
    state: str
    started_at: str
    completed_at: str
    message: str


@all_fields_optional
class ServiceInspectResult(DockerCamelModel):
    id: str = Field(alias="ID")
    version: ServiceVersion
    created_at: datetime
    updated_at: datetime
    spec: ServiceSpec
    previous_spec: ServiceSpec
    endpoint: ServiceEndpoint
    update_status: ServiceUpdateStatus



import base64
import json
from typing import Dict, List, Optional

from .client_config import ClientConfig, DockerCLICaller
from .components.buildx.cli_wrapper import BuildxCLI
from .components.compose.cli_wrapper import ComposeCLI
from .components.config.cli_wrapper import ConfigCLI
from .components.container.cli_wrapper import ContainerCLI
from .components.context.cli_wrapper import ContextCLI
from .components.image.cli_wrapper import ImageCLI
from .components.manifest.cli_wrapper import ManifestCLI
from .components.network.cli_wrapper import NetworkCLI
from .components.node.cli_wrapper import NodeCLI
from .components.plugin.cli_wrapper import PluginCLI
from .components.secret.cli_wrapper import SecretCLI
from .components.service.cli_wrapper import ServiceCLI
from .components.stack.cli_wrapper import StackCLI
from .components.swarm.cli_wrapper import SwarmCLI
from .components.system.cli_wrapper import SystemCLI
from .components.task.cli_wrapper import TaskCLI
from .components.trust.cli_wrapper import TrustCLI
from .components.volume.cli_wrapper import VolumeCLI
from .utils import ValidPath, run


class DockerClient(DockerCLICaller):
    """Creates a Docker client

    Note that
    ```python
    from dockertown import docker
    print(docker.run("hello-world"))
    ```
    is equivalent to
    ```python
    from dockertown import DockerClient
    docker = DockerClient()
    print(docker.run("hello-world")
    ```

    # Arguments
        config: Location of client config files (default "~/.docker")
        context: Name of the context to use to connect to the
            daemon (overrides DOCKER_HOST env var
            and default context set with "docker context use")
        debug: Enable debug mode
        host: Daemon socket(s) to connect to
        log_level: Set the logging level ("debug"|"info"|"warn"|"error"|"fatal")
           (default "info")
        tls:  Use TLS; implied by `tlsverify`
        tlscacert: Trust certs signed only by this CA (default "~/.docker/ca.pem")
        compose_files: Docker compose yaml file
        compose_profiles: List of compose profiles to use. Take a look at
            the [documentation for profiles](https://docs.docker.com/compose/profiles/).
        compose_env_file: .env file containing the environments variables to inject
            into the compose project. By default, it uses `./.env`.
        compose_project_name: The name of the compose project. It will be prefixed to
            networks, volumes and containers created by compose.
        compose_project_directory: Use an alternate working directory. By default, it
            uses the path of the compose file.
        compose_compatibility: Use docker compose in compatibility mode.
        client_call: Client binary to use and how to call it. Default is `["docker"]`. You can try with
            for example `["podman"]` or `["nerdctl"]`. The client must have the same commands and
            outputs as Docker to work. Some best effort support is done in case of divergences,
            meaning you can report issues occuring on some other binary than Docker, but
            we don't guarantee that it will be fixed.
            This option is a list because you can provide a list of command line arguments to be placed after `"docker"`.
            For exemple `host="ssh://my_user@host.com"` is equivalent
            to `client_call=["docker", "--host=ssh://my_user@host.com"]`.
            This will allow you to use some exotic options that are not explicitly supported by `dockertown`.
            Let's say you want to use estargz to run a container immediately, without waiting for the "pull"
            to finish (yes it's possible!), you can
            do `nerdctl = DockerClient(client_call=["nerdctl", "--snapshotter=stargz"])`
            and then `nerdctl.run("ghcr.io/stargz-containers/python:3.7-org", ["-c", "print('hi')"])`.
            You can also use this system to call Docker with sudo with `client_call=["sudo", "docker"]`
            (note that it won't ask for your password, so sudo should be passwordless during the python
            program execution).
        client_binary: Deprecated, use `client_call`. If you used before `client_binary="podman"`, now use
            `client_call=["podman"]`.
    """

    def __init__(
        self,
        config: Optional[ValidPath] = None,
        context: Optional[str] = None,
        debug: Optional[bool] = None,
        host: Optional[str] = None,
        log_level: Optional[str] = None,
        tls: Optional[bool] = None,
        tlscacert: Optional[ValidPath] = None,
        tlscert: Optional[ValidPath] = None,
        tlskey: Optional[ValidPath] = None,
        tlsverify: Optional[bool] = None,
        client_config: Optional[ClientConfig] = None,
        compose_files: List[ValidPath] = [],
        compose_profiles: List[str] = [],
        compose_env_file: Optional[ValidPath] = None,
        compose_project_name: Optional[str] = None,
        compose_project_directory: Optional[ValidPath] = None,
        compose_compatibility: Optional[bool] = None,
        client_call: List[str] = ["docker"],
    ):
        if client_config is None:
            client_config = ClientConfig(
                config=config,
                context=context,
                debug=debug,
                host=host,
                log_level=log_level,
                tls=tls,
                tlscacert=tlscacert,
                tlscert=tlscert,
                tlskey=tlskey,
                tlsverify=tlsverify,
                compose_files=compose_files,
                compose_profiles=compose_profiles,
                compose_env_file=compose_env_file,
                compose_project_name=compose_project_name,
                compose_project_directory=compose_project_directory,
                compose_compatibility=compose_compatibility,
                client_call=client_call,
            )
        super().__init__(client_config)

        self.buildx = BuildxCLI(self.client_config)
        self.compose = ComposeCLI(self.client_config)
        self.config = ConfigCLI(self.client_config)
        self.container = ContainerCLI(self.client_config)
        self.context = ContextCLI(self.client_config)
        self.image = ImageCLI(self.client_config)
        self.manifest = ManifestCLI(self.client_config)
        self.network = NetworkCLI(self.client_config)
        self.node = NodeCLI(self.client_config)
        self.plugin = PluginCLI(self.client_config)
        self.secret = SecretCLI(self.client_config)
        self.service = ServiceCLI(self.client_config)
        self.stack = StackCLI(self.client_config)
        self.swarm = SwarmCLI(self.client_config)
        self.system = SystemCLI(self.client_config)
        self.task = TaskCLI(self.client_config)
        self.trust = TrustCLI(self.client_config)
        self.volume = VolumeCLI(self.client_config)

        # aliases
        self.attach = self.container.attach
        self.build = self.buildx.build
        self.legacy_build = self.image.legacy_build
        self.commit = self.container.commit
        self.copy = self.container.copy
        self.create = self.container.create
        self.diff = self.container.diff
        self.events = None
        self.execute = self.container.execute
        self.export = self.container.export
        self.images = self.image.list
        self.import_ = self.image.import_
        self.info = self.system.info
        # self.inspect -> too hard to implement
        self.kill = self.container.kill
        self.load = self.image.load
        self.logs = self.container.logs
        self.pause = self.container.pause
        self.ps = self.container.list
        self.pull = self.image.pull
        self.push = self.image.push
        self.rename = self.container.rename
        self.restart = self.container.restart
        self.remove = self.container.remove
        # self.rmi -> doesn't make much sense since it would be docker.remove_image
        self.run = self.container.run
        self.save = self.image.save
        # self.search -> Is anybody going to use it in python?
        self.start = self.container.start
        self.stats = self.container.stats
        self.stop = self.container.stop
        self.tag = self.image.tag
        self.top = self.container.stop
        self.unpause = self.container.unpause
        self.update = self.container.update
        self.wait = self.container.wait

    def version(self) -> Dict:
        """
        Retrieves the docker client and engine versions.
        :return: dictionary containing the raw output of `docker version`
        """
        full_cmd = self.docker_cmd + ["version", "--format", "{{json .}}"]
        version_str = run(full_cmd, capture_stderr=False, capture_stdout=True)
        return json.loads(version_str)

    def login(
        self,
        server: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):
        """Log in to a Docker registry.

        If no server is specified, the default is defined by the daemon.

        # Arguments
            server: The server to log into. For example, with a self-hosted registry
                it might be something like `server="192.168.0.10:5000"`
            username: The username
            password: The password
        """
        full_cmd = self.docker_cmd + ["login"]

        full_cmd.add_simple_arg("--username", username)
        full_cmd.add_simple_arg("--password", password)
        if server is not None:
            full_cmd.append(server)

        run(full_cmd, capture_stderr=False, capture_stdout=False)

    def logout(self, server: Optional[str] = None):
        """Logout from a Docker registry

        # Arguments
            server: The server to logout from. For example, with a self-hosted registry
                it might be something like `server="192.168.0.10:5000"`
        """
        full_cmd = self.docker_cmd + ["logout"]

        if server is not None:
            full_cmd.append(server)

        run(full_cmd, capture_stdout=False, capture_stderr=False)

    def login_ecr(
        self,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        region_name: Optional[str] = None,
        registry: Optional[str] = None,
    ):
        """Login to the aws ECR registry. Credentials are taken from the
        environment variables as defined in
        [the aws docs](https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-envvars.html).

        If you don't have a profile or your environment variables configured, you can also
        use the function arguments `aws_access_key_id`, `aws_secret_access_key`, `region_name`.

        Behind the scenes, those arguments are passed directly to
        ```python
        botocore.session.get_session().create_client(...)
        ```

        You need botocore to run this function. Use `pip install botocore` to install it.

        The `registry` parameter can be used to override the registry that is guessed from the authorization token
        request's response.
        In other words: If the registry is `None` (the default) then it will be assumed that it's the ECR registry linked to the credentials provided.
        It is especially useful if the aws account you use can access several repositories and you
        need to explicitly define the one you want to use
        """
        import botocore.session

        client = botocore.session.get_session().create_client(
            "ecr",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        )

        response = client.get_authorization_token()["authorizationData"][0]
        credentials = base64.b64decode(response["authorizationToken"]).decode()
        username, password = credentials.split(":")
        if registry is None:
            registry = response["proxyEndpoint"]
        self.login(registry, username, password)



from typing import List

from dockertown.components.buildx.imagetools.models import ImageVariantManifest
from dockertown.utils import DockerCamelModel, all_fields_optional


@all_fields_optional
class ManifestListInspectResult(DockerCamelModel):
    name: str
    schema_version: int
    media_type: str
    manifests: List[ImageVariantManifest]



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, overload

from ...client_config import ClientConfig, DockerCLICaller, ReloadableObjectFromJson
from ...components.task import cli_wrapper as task_cli_wrapper
from ...utils import run, to_list
from .models import (
    NodeDescription,
    NodeInspectResult,
    NodeManagerStatus,
    NodeSpec,
    NodeStatus,
    NodeVersion,
)


class Node(ReloadableObjectFromJson):
    def __init__(
        self, client_config: ClientConfig, reference: str, is_immutable_id=False
    ):
        super().__init__(client_config, "id", reference, is_immutable_id)

    def _fetch_inspect_result_json(self, reference):
        return run(self.docker_cmd + ["node", "inspect", reference])

    def _parse_json_object(self, json_object: Dict[str, Any]) -> NodeInspectResult:
        return NodeInspectResult.parse_obj(json_object)

    def _get_inspect_result(self) -> NodeInspectResult:
        """Only there to allow tools to know the return type"""
        return super()._get_inspect_result()

    @property
    def id(self) -> str:
        return self._get_immutable_id()

    @property
    def version(self) -> NodeVersion:
        return self._get_inspect_result().version

    @property
    def created_at(self) -> datetime:
        return self._get_inspect_result().created_at

    @property
    def updated_at(self) -> datetime:
        return self._get_inspect_result().updated_at

    @property
    def spec(self) -> NodeSpec:
        return self._get_inspect_result().spec

    @property
    def description(self) -> NodeDescription:
        return self._get_inspect_result().description

    @property
    def status(self) -> NodeStatus:
        return self._get_inspect_result().status

    @property
    def manager_status(self) -> Optional[NodeManagerStatus]:
        return self._get_inspect_result().manager_status

    def __repr__(self):
        return f"dockertown.Node(id='{self.id[:12]}', hostname='{self.description.hostname}', role='{self.spec.role}')"

    def update(
        self,
        availability: Optional[str] = None,
        labels_add: Dict[str, str] = {},
        rm_labels: List[str] = [],
        role: Optional[str] = None,
    ) -> None:
        """Updates this Swarm node.

        See [`docker.node.update`](../sub-commands/node.md#update) for more information
        about the arguments.
        """
        NodeCLI(self.client_config).update(
            self, availability, labels_add, rm_labels, role
        )

    def ps(self) -> List[task_cli_wrapper.Task]:
        """Returns the list of tasks running on this node

        # Returns
            A `List[dockertown.Task]` object.

        """
        return NodeCLI(self.client_config).ps(self)


ValidNode = Union[Node, str]


class NodeCLI(DockerCLICaller):
    def demote(self, x: Union[ValidNode, List[ValidNode]]):
        """Demote one or more nodes from manager in the swarm

        # Arguments
            x: One or a list of nodes.
        """
        full_cmd = self.docker_cmd + ["node", "demote"]
        if x == []:
            return
        full_cmd += to_list(x)
        run(full_cmd)

    @overload
    def inspect(self, x: str) -> Node:
        ...

    @overload
    def inspect(self, x: List[str]) -> List[Node]:
        ...

    def inspect(self, x: Union[str, List[str]]) -> Union[Node, List[Node]]:
        """Returns a `dockertown.Node` object from a string
        (id or hostname of the node)

        # Arguments
            x: One id or hostname or a list of ids or hostnames

        # Returns
            One or a list of `dockertown.Node`
        """
        if isinstance(x, str):
            return Node(self.client_config, x)
        else:
            return [Node(self.client_config, reference) for reference in x]

    def list(self) -> List[Node]:
        """Returns the list of nodes in this swarm.

        # Returns
            A `List[dockertown.Node]`
        """
        full_cmd = self.docker_cmd + ["node", "list", "--quiet"]
        all_ids = run(full_cmd).splitlines()
        return [Node(self.client_config, x, is_immutable_id=True) for x in all_ids]

    def promote(self, x: Union[ValidNode, List[ValidNode]]):
        """Promote one or more nodes to manager in the swarm

        # Arguments
            x: One or a list of nodes.
        """
        full_cmd = self.docker_cmd + ["node", "promote"]
        if x == []:
            return
        full_cmd += to_list(x)
        run(full_cmd)

    def ps(
        self, x: Union[ValidNode, List[ValidNode], None] = None
    ) -> List[task_cli_wrapper.Task]:
        """Returns the list of swarm tasks running on one or more nodes.

        ```python
        from dockertown import docker

        tasks = docker.node.ps("my-node-name")
        print(tasks[0].desired_state)
        # running
        ```

        # Arguments
            x: One or more nodes (can be id, name or `dockertown.Node` object.).
                If the argument is not provided, it defaults to the current node.
                An empty list means an empty list will also be returned.

        # Returns
            `List[dockertown.Task]`
        """
        if x == []:
            return []
        elif x is None:
            # this is the default, on the command line, nothing is provided, and then
            # it automatically defaults to the current node.
            command_args = []
        else:
            command_args = to_list(x)
        full_cmd = (
            self.docker_cmd + ["node", "ps", "--quiet", "--no-trunc"] + command_args
        )
        ids = run(full_cmd).splitlines()
        return [
            task_cli_wrapper.Task(self.client_config, id_, is_immutable_id=True)
            for id_ in ids
        ]

    def remove(self, x: Union[ValidNode, List[ValidNode]], force: bool = False):
        """Remove one or more nodes from the swarm

        # Arguments
            x: One node or a list of nodes. You can use the id or the hostname of a node.
                You can also use a `dockertown.Node`.
            force: Force remove a node from the swarm
        """
        full_cmd = self.docker_cmd + ["node", "remove"]
        full_cmd.add_flag("--force", force)
        if x == []:
            return
        full_cmd += to_list(x)
        run(full_cmd)

    def update(
        self,
        node: ValidNode,
        availability: Optional[str] = None,
        labels_add: Dict[str, str] = {},
        rm_labels: List[str] = [],
        role: Optional[str] = None,
    ) -> None:
        """Updates a Swarm node.

        # Arguments
            node: The node to update, you can use a string or a `dockertown.Node`
                object.
            availability: Availability of the node ("active"|"pause"|"drain")
            labels_add: Remove a node label if exists
            rm_labels: Labels to remove from the node.
            role: Role of the node ("worker"|"manager")
        """
        full_cmd = self.docker_cmd + ["node", "update"]

        full_cmd.add_simple_arg("--availability", availability)
        for label_name, label_value in labels_add.items():
            full_cmd += ["--label-add", f"{label_name}={label_value}"]

        for label in rm_labels:
            full_cmd += ["--label-rm", label]

        full_cmd.add_simple_arg("--role", role)
        full_cmd.append(node)
        run(full_cmd)



import datetime
import json
from typing import Dict, Iterator, Union

from ...client_config import DockerCLICaller
from ...utils import format_dict_for_cli, format_time_arg, run, stream_stdout_and_stderr
from .models import DockerEvent, DockerItemsSummary, SystemInfo


class DiskFreeResult:
    def __init__(self, cli_stdout: str):
        docker_items = {}
        for line in cli_stdout.splitlines():
            docker_items_dict = json.loads(line)
            reclamable = docker_items_dict["Reclaimable"]
            docker_items_dict["Reclaimable"] = reclamable.split(" ")[0]
            if "%" in reclamable:
                docker_items_dict["ReclaimablePercent"] = reclamable.split(" ")[1][1:-2]
            else:
                docker_items_dict["ReclaimablePercent"] = "100"

            docker_items[docker_items_dict["Type"]] = docker_items_dict

        self.images: DockerItemsSummary
        self.images = DockerItemsSummary.parse_obj(docker_items["Images"])
        self.containers: DockerItemsSummary
        self.containers = DockerItemsSummary.parse_obj(docker_items["Containers"])
        self.volumes: DockerItemsSummary
        self.volumes = DockerItemsSummary.parse_obj(docker_items["Local Volumes"])
        self.build_cache: DockerItemsSummary
        self.build_cache = DockerItemsSummary.parse_obj(docker_items["Build Cache"])


class SystemCLI(DockerCLICaller):
    def disk_free(self) -> DiskFreeResult:
        """Give information about the disk usage of the Docker daemon.

        Returns a `dockertown.DiskFreeResult` object.

        ```python
        from dockertown import docker
        disk_free_result = docker.system.disk_free()
        print(disk_free_result.images.active)  #int
        print(disk_free_result.containers.reclaimable)  # int, number of bytes
        print(disk_free_result.volumes.reclaimable_percent)  # float
        print(disk_free_result.build_cache.total_count)  # int
        print(disk_free_result.build_cache.size)  # int, number of bytes
        ...
        ```
        Note that the number are not 100% accurate because the docker CLI
        doesn't provide the exact numbers.

        Maybe in a future implementation, we can provide exact numbers.

        Verbose mode is not yet implemented.
        """

        full_cmd = self.docker_cmd + ["system", "df", "--format", "{{json .}}"]
        return DiskFreeResult(run(full_cmd))

    def events(
        self,
        since: Union[None, datetime.datetime, datetime.timedelta] = None,
        until: Union[None, datetime.datetime, datetime.timedelta] = None,
        filters: Dict[str, str] = {},
    ) -> Iterator[DockerEvent]:
        """Return docker events information up to the current point in time.

        If `until` is not specified, then the iterator returned is infinite.
        For example

        ```python
        from dockertown import docker
        from datetime import datetime, timedelta


        for event in docker.system.events():
            print("new event!")
            print(event)
            # this will never end, that's ok if you want to monitor something
            # for a long time. You can also use 'break' in the for loop if needed.

        for event in docker.system.events(until=datetime.now() - timedelta(seconds=30)):
            print("some past event")
            print(event)
            # this loop will end, unlike the previous one

        for event in docker.system.events(until=datetime.now() + timedelta(seconds=30)):
            print("some past event")
            print(event)
            # this loop will end in 30 seconds, even if there are no events at all

        events = list(docker.system.events(filters={"container": "mycontainer"}, until=datetime.now()))
        # the list of all events concerning the container "mycontainer"
        ```

        # Arguments
            since:  Show all events created since timestamp
            until: 	Stream events until this timestamp
            filters: See the [Docker documentation page about filtering
                ](https://docs.docker.com/engine/reference/commandline/events/#filtering).
        # Returns
            A iterator which will yield DockerEvent objects from stdout/stderr

        [reference page for
        system events](https://docs.docker.com/engine/api/v1.40/#operation/SystemEvents)
        """
        full_cmd = self.docker_cmd + [
            "system",
            "events",
            "--format",
            "{{json .}}",
        ]
        full_cmd.add_simple_arg("--since", format_time_arg(since))
        full_cmd.add_simple_arg("--until", format_time_arg(until))
        full_cmd.add_args_list("--filter", format_dict_for_cli(filters))
        iterator = stream_stdout_and_stderr(full_cmd)
        for stream_origin, stream_content in iterator:
            if stream_origin == "stdout":
                yield DockerEvent.parse_raw(stream_content)

    def info(self) -> SystemInfo:
        """Returns diverse information about the Docker client and daemon.

        # Returns
            A `dockertown.SystemInfo` object

        As an example

        ```python
        from dockertown import docker

        info = docker.system.info()
        print(info.images)
        # 40
        print(info.plugins.volume)
        # ["local"}
        ...
        ```

        You can find all attributes available by looking up the [reference page for
        system info](https://docs.docker.com/engine/api/v1.40/#operation/SystemInfo).
        """
        full_cmd = self.docker_cmd + ["system", "info", "--format", "{{json .}}"]
        return SystemInfo.parse_raw(run(full_cmd))

    def prune(
        self, all: bool = False, volumes: bool = False, filters: Dict[str, str] = {}
    ) -> None:
        """Remove unused docker data

        # Arguments
            all: Remove all unused images not just dangling ones
            volumes: Prune volumes
            filters: See the [Docker documentation page about filtering
                ](https://docs.docker.com/engine/reference/commandline/system_prune/#filtering).
                For example, `filters=dict(until="24h")`.
        """
        full_cmd = self.docker_cmd + ["system", "prune", "--force"]
        full_cmd.add_flag("--all", all)
        full_cmd.add_flag("--volumes", volumes)
        full_cmd.add_args_list("--filter", format_dict_for_cli(filters))
        run(full_cmd)



from typing import List

from ...utils import DockerCamelModel, all_fields_optional


@all_fields_optional
class PluginMount(DockerCamelModel):
    pass


@all_fields_optional
class PluginDevice(DockerCamelModel):
    pass


@all_fields_optional
class PluginSettings(DockerCamelModel):
    mounts: List[PluginMount]
    env: List[str]
    args: List[str]
    devices: List[PluginDevice]


@all_fields_optional
class Interface(DockerCamelModel):
    pass


@all_fields_optional
class PluginConfig(DockerCamelModel):
    docker_version: str
    description: str
    documentation: str
    interface: Interface
    entrypoint: List[str]
    work_dir: str
    # TODO: add missing attributes


@all_fields_optional
class PluginInspectResult(DockerCamelModel):
    id: str
    name: str
    enabled: bool
    settings: PluginSettings
    plugin_reference: str
    config: PluginConfig



from typing import Any, Dict, List, Optional, Union

from ...client_config import ClientConfig, DockerCLICaller, ReloadableObjectFromJson
from ...utils import run, to_list
from ..buildx.imagetools.models import ImageVariantManifest
from .models import ManifestListInspectResult


class ManifestList(ReloadableObjectFromJson):
    def __init__(
        self, client_config: ClientConfig, reference: str, is_immutable_id=False
    ):
        self.reference = reference
        super().__init__(client_config, "name", reference, is_immutable_id)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove()

    def _fetch_inspect_result_json(self, reference):
        return f'[{run(self.docker_cmd + ["manifest", "inspect", reference])}]'

    def _parse_json_object(
        self, json_object: Dict[str, Any]
    ) -> ManifestListInspectResult:
        json_object["name"] = self.reference
        return ManifestListInspectResult.parse_obj(json_object)

    def _get_inspect_result(self) -> ManifestListInspectResult:
        """Only there to allow tools to know the return type"""
        return super()._get_inspect_result()

    @property
    def name(self) -> str:
        return self._get_inspect_result().name

    @property
    def schema_version(self) -> str:
        return self._get_inspect_result().schema_version

    @property
    def media_type(self) -> str:
        return self._get_inspect_result().media_type

    @property
    def manifests(self) -> List[ImageVariantManifest]:
        return self._get_inspect_result().manifests

    def __repr__(self):
        return f"dockertown.ManifestList(name='{self.name}')"

    def remove(self) -> None:
        """Removes this Docker manifest list.

        Rather than removing it manually, you can use a context manager to
        make sure the manifest list is deleted even if an exception is raised.
        """
        ManifestCLI(self.client_config).remove(self)


ValidManifestList = Union[ManifestList, str]


class ManifestCLI(DockerCLICaller):
    def annotate(
        self,
        name: str,
        manifest: str,
        arch: Optional[str] = None,
        os: Optional[str] = None,
        os_features: Optional[List[str]] = None,
        os_version: Optional[str] = None,
        variant: Optional[str] = None,
    ) -> ManifestList:
        """Annotates a Docker manifest list.

        # Arguments
            name: The name of the manifest list
            manifest: The individual manifest to annotate
            arch: The manifest's architecture
            os: The manifest's operating system
            os_features: The manifest's operating system features
            os_version: The manifest's operating system version
            variant: The manifest's architecture variant
        """
        full_cmd = self.docker_cmd + ["manifest", "annotate"]
        full_cmd.add_simple_arg("--arch", arch)
        full_cmd.add_simple_arg("--os", os)
        full_cmd.add_simple_arg("--os-features", os_features)
        full_cmd.add_simple_arg("--os-version", os_version)
        full_cmd.add_simple_arg("--variant", variant)
        full_cmd.append(name)
        full_cmd.append(manifest)
        run(full_cmd)

    def create(
        self,
        name: str,
        manifests: List[str],
        ammend: bool = False,
        insecure: bool = False,
    ) -> ManifestList:
        """Creates a Docker manifest list.

        # Arguments
            name: The name of the manifest list
            manifests: The list of manifests to add to the manifest list

        # Returns
            A `dockertown.ManifestList`.
        """
        full_cmd = self.docker_cmd + ["manifest", "create"]
        full_cmd.add_flag("--amend", ammend)
        full_cmd.add_flag("--insecure", insecure)
        full_cmd.append(name)
        full_cmd += to_list(manifests)
        return ManifestList(
            self.client_config, run(full_cmd)[22:], is_immutable_id=True
        )

    def inspect(self, x: str) -> ManifestList:
        """Returns a Docker manifest list object."""
        return ManifestList(self.client_config, x)

    def push(self, x: str, purge: bool = False, quiet: bool = False):
        """Push a manifest list to a repository.

        # Options
            purge: Remove the local manifest list after push
        """
        # this is just to raise a correct exception if the manifest list doesn't exist
        self.inspect(x)

        full_cmd = self.docker_cmd + ["manifest", "push"]
        full_cmd.add_flag("--purge", purge)
        full_cmd.append(x)
        run(full_cmd, capture_stdout=quiet, capture_stderr=quiet)

    def remove(self, manifest_lists: Union[ValidManifestList, List[ValidManifestList]]):
        """Removes a Docker manifest list or lists.

        # Arguments
            manifest_lists: One or more manifest lists.
        """
        if manifest_lists == []:
            return
        full_cmd = self.docker_cmd + ["manifest", "rm"]
        full_cmd += to_list(manifest_lists)
        run(full_cmd)



from pathlib import Path

from setuptools import find_packages, setup

CURRENT_DIR = Path(__file__).parent


def get_long_description() -> str:
    return (CURRENT_DIR / "README.md").read_text(encoding="utf8")


setup(
    name="dockertown",
    version="0.2.2",
    description="A decent Python wrapper for Docker CLI",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    install_requires=(CURRENT_DIR / "requirements.txt").read_text().splitlines(),
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,  # will read the MANIFEST.in
    license="MIT",
    python_requires=">=3.7, <4",
    entry_points={
        "console_scripts": ["dockertown=dockertown.command_line_entrypoint:main"],
    },
    project_urls={
        "Documentation": "https://duckietown.github.io/dockertown/",
        "Source Code": "https://github.com/duckietown/dockertown",
        "Bug Tracker": "https://github.com/duckietown/dockertown/issues",
    },
)



__version__ = "0.2.2"

from .client_config import ClientNotFoundError
from .components.buildx.cli_wrapper import Builder
from .components.config.cli_wrapper import Config
from .components.container.cli_wrapper import Container, ContainerStats
from .components.context.cli_wrapper import (
    Context,
    DockerContextConfig,
    KubernetesContextConfig,
)
from .components.image.cli_wrapper import Image
from .components.network.cli_wrapper import Network
from .components.node.cli_wrapper import Node
from .components.plugin.cli_wrapper import Plugin
from .components.secret.cli_wrapper import Secret
from .components.service.cli_wrapper import Service
from .components.stack.cli_wrapper import Stack
from .components.system.cli_wrapper import SystemInfo
from .components.task.cli_wrapper import Task
from .components.volume.cli_wrapper import Volume
from .docker_client import DockerClient
from .exceptions import DockerException

# alias
docker = DockerClient()

__all__ = [
    "Builder",
    "ClientNotFoundError",
    "Config",
    "Container",
    "ContainerStats",
    "Context",
    "DockerContextConfig",
    "DockerException",
    "Image",
    "KubernetesContextConfig",
    "Network",
    "Node",
    "Plugin",
    "Secret",
    "Service",
    "Stack",
    "SystemInfo",
    "Task",
    "Volume",
    "DockerClient"
]



from datetime import datetime
from typing import Any, Dict, List

from ...utils import DockerCamelModel, all_fields_optional


@all_fields_optional
class NetworkIPAM(DockerCamelModel):
    driver: str
    config: List[Dict[str, Any]]
    options: Dict[str, Any]


@all_fields_optional
class NetworkContainer(DockerCamelModel):
    name: str
    endpoint_id: str
    mac_address: str
    ipv4_address: str
    ipv6_address: str


@all_fields_optional
class NetworkInspectResult(DockerCamelModel):
    name: str
    id: str
    created: datetime
    scope: str
    driver: str
    enable_ipv6: bool
    ipam: NetworkIPAM
    internal: bool
    attachable: bool
    ingress: bool
    containers: Dict[str, NetworkContainer]
    options: Dict[str, Any]
    labels: Dict[str, str]
    config_from: dict
    config_only: bool



from __future__ import annotations

import os
import pty
import re
import subprocess
from datetime import datetime
from multiprocessing.pool import ThreadPool
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Any, Dict, Iterator, List, Optional, Union, overload, Match

import humanfriendly

from ...client_config import ClientConfig, DockerCLICaller, ReloadableObjectFromJson
from ...components.buildx import cli_wrapper as buildx_cli_wrapper
from ...components.container import cli_wrapper as container_cli_wrapper
from ...exceptions import DockerException, NoSuchImage
from ...utils import (
    ValidPath,
    format_dict_for_cli,
    run,
    stream_stdout_and_stderr,
    to_list,
)
from .models import ImageGraphDriver, ImageInspectResult, ImageRootFS, ImageHistoryLayer, LayerPullStatus, \
    LayerPullProgress


class Image(ReloadableObjectFromJson):
    def __init__(
        self, client_config: ClientConfig, reference: str, is_immutable_id=False
    ):
        super().__init__(client_config, "id", reference, is_immutable_id)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove(force=True)

    def _fetch_inspect_result_json(self, reference):
        return run(self.docker_cmd + ["image", "inspect", reference])

    def _parse_json_object(self, json_object: Dict[str, Any]) -> ImageInspectResult:
        return ImageInspectResult.parse_obj(json_object)

    def _get_inspect_result(self) -> ImageInspectResult:
        """Only there to allow tools to know the return type"""
        return super()._get_inspect_result()

    @property
    def id(self) -> str:
        return self._get_immutable_id()

    @property
    def repo_tags(self) -> List[str]:
        return self._get_inspect_result().repo_tags

    @property
    def repo_digests(self) -> List[str]:
        return self._get_inspect_result().repo_digests

    @property
    def parent(self) -> str:
        return self._get_inspect_result().parent

    @property
    def comment(self) -> str:
        return self._get_inspect_result().comment

    @property
    def created(self) -> datetime:
        return self._get_inspect_result().created

    @property
    def container(self) -> str:
        return self._get_inspect_result().container

    @property
    def container_config(
        self,
    ) -> container_cli_wrapper.ContainerConfig:
        return self._get_inspect_result().container_config

    @property
    def docker_version(self) -> str:
        return self._get_inspect_result().docker_version

    @property
    def author(self) -> str:
        return self._get_inspect_result().author

    @property
    def config(
        self,
    ) -> container_cli_wrapper.ContainerConfig:
        return self._get_inspect_result().config

    @property
    def architecture(self) -> str:
        return self._get_inspect_result().architecture

    @property
    def os(self) -> str:
        return self._get_inspect_result().os

    @property
    def os_version(self) -> str:
        return self._get_inspect_result().os_version

    @property
    def size(self) -> int:
        return self._get_inspect_result().size

    @property
    def virtual_size(self) -> int:
        return self._get_inspect_result().virtual_size

    @property
    def graph_driver(self) -> ImageGraphDriver:
        return self._get_inspect_result().graph_driver

    @property
    def root_fs(self) -> ImageRootFS:
        return self._get_inspect_result().root_fs

    @property
    def metadata(self) -> Optional[Dict[str, str]]:
        return self._get_inspect_result().metadata

    def __repr__(self):
        return f"dockertown.Image(id='{self.id[:20]}', tags={self.repo_tags})"

    def remove(self, force: bool = False, prune: bool = True):
        """Remove this Docker image.

        See the [`docker.image.remove`](../sub-commands/container.md#remove) command for
        information about the arguments.
        """
        ImageCLI(self.client_config).remove(self, force, prune)

    def save(self, output: Optional[ValidPath] = None) -> Optional[Iterator[bytes]]:
        """Saves this Docker image in a tar.

        See the [`docker.image.save`](../sub-commands/container.md#save) command for
        information about the arguments.
        """
        return ImageCLI(self.client_config).save(self, output)

    def tag(self, new_tag: str) -> None:
        """Add a tag to a Docker image.

        See the [`docker.image.tag`](../sub-commands/container.md#tag) command for
        information about the arguments.
        """
        return ImageCLI(self.client_config).tag(self, new_tag)

    def copy_from(
        self, path_in_image: ValidPath, destination: ValidPath, pull: str = "missing"
    ):
        """Copy a file from a docker image in the local filesystem.

        See the `docker.image.copy_from` command for information about the arguments.
        """
        return ImageCLI(self.client_config).copy_from(
            self, path_in_image, destination, pull
        )

    def copy_to(
        self,
        local_path: ValidPath,
        path_in_image: ValidPath,
        new_tag: Optional[str] = None,
        pull: str = "missing",
    ) -> Image:
        """Copy a file from the local filesystem in a docker image to create a new
        docker image.

        As if you did a dockerfile with a COPY instruction.

        See the `docker.image.copy_to` command for information about the arguments.
        """
        return ImageCLI(self.client_config).copy_to(
            self, local_path, path_in_image, new_tag, pull
        )

    def exists(self) -> bool:
        """Returns `True` if the docker image exists and `False` if it doesn't exists.

        Note that you might have done `docker.image.remove("some_tag")` and the image
        might still exists because dockertown references images by id, not by tag.

         See the `docker.image.exists` command for information about the arguments.
        """
        return ImageCLI(self.client_config).exists(self.id)


ValidImage = Union[str, Image]


class ImageCLI(DockerCLICaller):
    def __init__(self, client_config: ClientConfig):
        super().__init__(client_config)
        self.build = buildx_cli_wrapper.BuildxCLI(client_config).build

    def legacy_build(
        self,
        context_path: ValidPath,
        add_hosts: Dict[str, str] = {},
        build_args: Dict[str, str] = {},
        cache: bool = True,
        file: Optional[ValidPath] = None,
        labels: Dict[str, str] = {},
        network: Optional[str] = None,
        pull: bool = False,
        tags: Union[str, List[str]] = [],
        target: Optional[str] = None,
    ) -> Image:
        """Build a Docker image with the old Docker builder (meaning not using buildx/buildkit)

        As the name implies this is a legacy building method. Users are strongly encouraged to use
        `docker.build()` instead. The legacy builder will not be available in docker v22.06 and above.

        This function also won't run the legacy builder if the environment variable
        `DOCKER_BUILDKIT` is set to `1` or if you had run previously `docker buildx install` from bash
        or `docker.buildx.install()` from Python.

        Some resources on why moving to buildx/buildkit is necessary:

        * [Proposal: make BuildKit the default builder on Linux](https://github.com/moby/moby/issues/40379)
        * [Deprecated Engine Features: Legacy builder for Linux images](https://github.com/docker/cli/blob/master/docs/deprecated.md#legacy-builder-for-linux-images)

        A `dockertown.Image` is returned, even when using multiple tags.
        That is because it will produce a single image with multiple tags.

        # Arguments
            context_path: The path of the build context. Defaults to the current working directory
            add_hosts: Hosts to add. `add_hosts={"my_host1": "192.168.32.35"}`
            build_args: The build arguments.
                ex `build_args={"PY_VERSION": "3.7.8", "UBUNTU_VERSION": "20.04"}`.
            cache: Whether or not to use the cache, defaults to True
            file: The path of the Dockerfile, defaults to `context_path/Dockerfile`
            labels: Dict of labels to add to the image.
                `labels={"very-secure": "1", "needs-gpu": "0"}` for example.
            network: which network to use when building the Docker image
            pull: Always attempt to pull a newer version of the image
            tags: Tag or tags to put on the resulting image.
            target: Set the target build stage to build.

        # Returns
            A `dockertown.Image`
        """
        # to make it easier to write and read tests, the tests of this function
        # are also grouped with the tests of "docker.build()".
        tags = to_list(tags)
        full_cmd = self.docker_cmd + ["build", "--quiet"]

        full_cmd.add_args_list(
            "--add-host", format_dict_for_cli(add_hosts, separator=":")
        )
        full_cmd.add_args_list("--build-arg", format_dict_for_cli(build_args))
        full_cmd.add_args_list("--label", format_dict_for_cli(labels))
        full_cmd.add_flag("--pull", pull)
        full_cmd.add_simple_arg("--file", file)
        full_cmd.add_simple_arg("--target", target)
        full_cmd.add_simple_arg("--network", network)
        full_cmd.add_flag("--no-cache", not cache)
        full_cmd.add_args_list("--tag", tags)

        docker_image = ImageCLI(self.client_config)
        full_cmd.append(context_path)
        image_id = run(full_cmd).splitlines()[-1].strip()
        return docker_image.inspect(image_id)

    def history(self, x: str) -> List[ImageHistoryLayer]:
        """Not yet implemented"""
        full_cmd = self.docker_cmd + ["history", "--no-trunc", "--format", "{{json .}}", x]
        history: List[ImageHistoryLayer] = [
            ImageHistoryLayer.parse_json(layer)
            for layer in run(full_cmd).splitlines()
        ]
        return history

    def import_(
        self,
        source: ValidPath,
        tag: Optional[str] = None,
        changes: List[str] = [],
        message: Optional[str] = None,
        platform: Optional[str] = None,
    ) -> Image:
        """Import the contents from a tarball to create a filesystem image

        Alias: `docker.import_(...)`

        # Arguments
            changes: Apply Dockerfile instruction to the created image
            message: Set commit message for imported image
            platform: Set platform if server is multi-platform capable
        """
        full_cmd = self.docker_cmd + ["image", "import"]
        full_cmd.add_args_list("--change", changes)
        full_cmd.add_simple_arg("--message", message)
        full_cmd.add_simple_arg("--platform", platform)
        full_cmd.append(source)
        if tag is not None:
            full_cmd.append(tag)
        return Image(self.client_config, run(full_cmd))

    @overload
    def inspect(self, x: str) -> Image:
        ...

    @overload
    def inspect(self, x: List[str]) -> List[Image]:
        ...

    def inspect(self, x: Union[str, List[str]]) -> Union[Image, List[Image]]:
        """Creates a `dockertown.Image` object.

        # Returns
            `dockertown.Image`, or `List[dockertown.Image]` if the input
            was a list of strings.

        # Raises
            `dockertown.exceptions.NoSuchImage` if one of the images does not exists.

        """
        if isinstance(x, list):
            return [Image(self.client_config, identifier) for identifier in x]
        else:
            return Image(self.client_config, x)

    def exists(self, x: str) -> bool:
        """Returns `True` if the image exists. `False` otherwise.

         It's just calling `docker.image.inspect(...)` and verifies that it doesn't throw
         a `dockertown.exceptions.NoSuchImage`.

        # Returns
            A `bool`
        """
        try:
            self.inspect(x)
        except NoSuchImage:
            return False
        else:
            return True

    def load(
        self, input: Union[ValidPath, bytes, Iterator[bytes]], quiet: bool = False
    ) -> List[str]:
        """Loads one or multiple Docker image(s) from a tar or an iterator of `bytes`.

        Alias: `docker.load(...)`

        # Arguments
            input: Path or input stream to load the images from.
            quiet: If you don't want to display the progress bars.

        # Returns
            `None` when using bytes as input. A `List[str]` (list of tags)
             when a path is provided.
        """
        full_cmd = self.docker_cmd + ["image", "load"]

        if isinstance(input, (str, Path)):
            full_cmd += ["--input", str(input)]

        if quiet:
            full_cmd.append("--quiet")

        if isinstance(input, (str, Path)):
            all_tags = []
            for source, stream_bytes in stream_stdout_and_stderr(full_cmd):
                decoded = stream_bytes.decode()[:-1]
                if not quiet:
                    print(decoded)
                if "Loaded image" in decoded:
                    all_tags.append(decoded.split(" ")[-1])
            return all_tags

        if isinstance(input, bytes):
            input = [input]

        stdout_lines = self._load_from_generator(full_cmd, input)
        all_tags = []
        for line in stdout_lines:
            if "Loaded image" in line:
                all_tags.append(line.split(" ")[-1])
        return all_tags

    def _load_from_generator(self, full_cmd: List[str], input: Iterator[bytes]):
        p = Popen(full_cmd, stdin=PIPE, stdout=PIPE)
        for buffer_bytes in input:
            p.stdin.write(buffer_bytes)
            p.stdin.flush()
        p.stdin.close()
        stdout = p.stdout.read()
        p.stdout.close()
        exit_code = p.wait()
        if exit_code != 0:
            raise DockerException(full_cmd, exit_code)
        return stdout.decode().splitlines()

    def list(
        self,
        repository_or_tag: Optional[str] = None,
        filters: Dict[str, str] = {},
        all: bool = False,
    ) -> List[Image]:
        """Returns the list of Docker images present on the machine.

        Alias: `docker.images()`

        Note that each image may have multiple tags.

        # Returns
            A `List[dockertown.Image]` object.
        """
        # previously the signature was
        # def list(self,filters: Dict[str, str] = {}) -> List[Image]:
        # so to avoid breakages when people used positional arguments, we can check the types and send a warning
        if isinstance(repository_or_tag, dict):
            filters = repository_or_tag
            repository_or_tag = None

        full_cmd = self.docker_cmd + [
            "image",
            "list",
            "--quiet",
            "--no-trunc",
        ]
        full_cmd.add_args_list("--filter", format_dict_for_cli(filters))
        full_cmd.add_flag("--all", all)

        if repository_or_tag is not None:
            full_cmd.append(repository_or_tag)

        ids = run(full_cmd).splitlines()
        # the list of tags is bigger than the number of images. We uniquify
        ids = set(ids)

        return [Image(self.client_config, x, is_immutable_id=True) for x in ids]

    def prune(self, all: bool = False, filter: Dict[str, str] = {}) -> str:
        """Remove unused images

        # Arguments
            all: Remove all unused images, not just dangling ones
            filter: Provide filter values (e.g. `{"until": "<timestamp>"}`)

        # Returns:
            The output of the CLI (the layers removed).
        """
        full_cmd = self.docker_cmd + ["image", "prune", "--force"]
        full_cmd.add_flag("--all", all)
        full_cmd.add_args_list("--filter", format_dict_for_cli(filter))
        return run(full_cmd)

    def pull(
        self, x: Union[str, List[str]], quiet: bool = False
    ) -> Union[Image, List[Image]]:
        """Pull one or more docker image(s)

        Alias: `docker.pull(...)`

        # Arguments
            x: The image name(s) . Can be a string or a list of strings. In case of
                list, multithreading is used to pull the images.
                The progress bars might look strange as multiple
                processes are drawing on the terminal at the same time.
            quiet: If you don't want to see the progress bars.

        # Returns:
            The Docker image loaded (`dockertown.Image` object).
            If a list was passed as input, then a `List[dockertown.Image]` will
            be returned.
        """

        if x == []:
            return []
        elif isinstance(x, str):
            return self._pull_single_tag(x, quiet=quiet)
        elif isinstance(x, list) and len(x) == 1:
            return [self._pull_single_tag(x[0], quiet=quiet)]
        elif len(x) >= 2:
            pool = ThreadPool(4)
            generator = self._generate_args_push_pull(x, quiet)
            all_images = pool.starmap(self._pull_single_tag, generator)
            pool.close()
            pool.join()
            return all_images

    def interactive_pull(self, x: str) -> Iterator[LayerPullProgress]:
        full_cmd = list(map(str, self.docker_cmd)) + ["pull", x]

        master, slave = pty.openpty()
        p = subprocess.Popen(full_cmd, preexec_fn=os.setsid, stdin=slave, stdout=slave, stderr=slave)
        os.close(slave)

        ansi_escape = re.compile(rb'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        layer_pattern = re.compile(rb"^([0-9a-f]{12}):\s(.*)$")
        progress_pattern = re.compile(r"^(\d+(\.\d+)?[kMGT]?B)/(\d+(\.\d+)?[kMGT]?B)$")

        buffer: bytes = b""

        while p.poll() is None:
            try:
                chunk = os.read(master, 2048)
            except OSError:
                return
            buffer += chunk

            if b"manifest unknown" in buffer:
                p.wait()
                raise NoSuchImage(full_cmd, p.returncode, stderr=buffer)

            buffer = ansi_escape.sub(b'', buffer)
            buffer = buffer.replace(b"\r", b"\n")
            lines = buffer.splitlines(keepends=True)

            for line in lines[:-1]:
                line = line.strip()
                if len(line) > 1:
                    # process line
                    match = layer_pattern.match(line)
                    if match:
                        layer_id: str = match.group(1).decode("utf-8").strip()
                        status_str: str = match.group(2).decode("utf-8").strip()
                        progress: Optional[float] = None
                        progress_info: Optional[Match] = progress_pattern.match(status_str.split(" ")[-1])
                        if progress_info:
                            progress_str: str = status_str.split(" ")[-1]
                            status_str = status_str[:-len(progress_str)].strip()
                            partial: float = humanfriendly.parse_size(progress_info.group(1))
                            total: float = humanfriendly.parse_size(progress_info.group(3))
                            progress: float = partial / total
                        status: LayerPullStatus = LayerPullStatus.from_string(status_str)

                        yield LayerPullProgress(
                            layer_id=layer_id,
                            status=status,
                            progress=progress,
                        )

            buffer = lines[-1]

    def _pull_single_tag(self, image_name: str, quiet: bool):
        full_cmd = self.docker_cmd + ["image", "pull"]

        if quiet:
            full_cmd.append("--quiet")

        full_cmd.append(image_name)
        run(full_cmd, capture_stdout=quiet, capture_stderr=quiet)
        return Image(self.client_config, image_name)

    def push(self, x: Union[str, List[str]], quiet: bool = False):
        """Push a tag or a repository to a registry

        Alias: `docker.push(...)`

        # Arguments
            x: Tag(s) or repo(s) to push. Can be a string or a list of strings.
                If it's a list of string, dockertown will push all the images with
                multiple threads. The progress bars might look strange as multiple
                processes are drawing on the terminal at the same time.
            quiet: If you don't want to see the progress bars.

        # Raises
            `dockertown.exceptions.NoSuchImage` if one of the images does not exists.
        """
        x = to_list(x)

        # this is just to raise a correct exception if the images don't exist
        self.inspect(x)

        if x == []:
            return
        elif len(x) == 1:
            self._push_single_tag(x[0], quiet=quiet)
        elif len(x) >= 2:
            pool = ThreadPool(4)
            generator = self._generate_args_push_pull(x, quiet)
            pool.starmap(self._push_single_tag, generator)
            pool.close()
            pool.join()

    def _generate_args_push_pull(self, _list: List[str], quiet: bool):
        for tag in _list:
            yield tag, quiet

    def _push_single_tag(self, tag_or_repo: str, quiet: bool):
        full_cmd = self.docker_cmd + ["image", "push"]

        full_cmd.append(tag_or_repo)
        run(full_cmd, capture_stdout=quiet, capture_stderr=quiet)

    def remove(
        self,
        x: Union[ValidImage, List[ValidImage]],
        force: bool = False,
        prune: bool = True,
    ):
        """Remove one or more docker images.

        # Arguments
            x: Single image or list of Docker images to remove. You can use tags or
                `dockertown.Image` objects.
            force: Force removal of the image
            prune: Delete untagged parents

        # Raises
            `dockertown.exceptions.NoSuchImage` if one of the images does not exists.

        """

        full_cmd = self.docker_cmd + ["image", "rm"]
        full_cmd.add_flag("--force", force)
        full_cmd.add_flag("--no-prune", not prune)
        if x == []:
            return
        for image in to_list(x):
            full_cmd.append(image)

        run(full_cmd)

    def save(
        self,
        images: Union[ValidImage, List[ValidImage]],
        output: Optional[ValidPath] = None,
    ) -> Optional[Iterator[bytes]]:
        """Save one or more images to a tar archive. Returns a stream if output is `None`

        Alias: `docker.save(...)`

        # Arguments
            images: Single docker image or list of docker images to save
            output: Path of the tar archive to produce. If `output` is None, a generator
                of bytes is produced. It can be used to stream those bytes elsewhere,
                to another Docker daemon for example.

        # Returns
            `Optional[Iterator[bytes]]`. If output is a path, nothing is returned.

        # Raises
            `dockertown.exceptions.NoSuchImage` if one of the images does not exists.

        # Example

        An example of transfer of an image from a local Docker daemon to a remote Docker
        daemon. We assume that the remote machine has an ssh access:

        ```python
        from dockertown import DockerClient

        local_docker = DockerClient()
        remote_docker = DockerClient(host="ssh://my_user@186.167.32.84")

        image_name = "busybox:1"
        local_docker.pull(image_name)
        bytes_iterator = local_docker.image.save(image_name)

        remote_docker.image.load(bytes_iterator)
        ```

        Of course the best solution is to use a registry to transfer image but
        it's a cool example nonetheless.
        """
        full_cmd = self.docker_cmd + ["image", "save"]
        images = to_list(images)

        # trigger an exception early
        self.inspect(images)

        if output is not None:
            full_cmd += ["--output", str(output)]

        full_cmd += images
        if output is None:
            # we stream the bytes
            return self._save_generator(full_cmd)
        else:
            run(full_cmd)

    def _save_generator(self, full_cmd) -> Iterator[bytes]:
        full_cmd = [str(x) for x in full_cmd]
        p = Popen(full_cmd, stdout=PIPE, stderr=PIPE)
        for line in p.stdout:
            yield line
        exit_code = p.wait(0.1)
        if exit_code != 0:
            stderr = p.stderr.read()
            if "No such image" in stderr.decode():
                raise NoSuchImage(full_cmd, exit_code, stderr=stderr)
            raise DockerException(full_cmd, exit_code, stderr=stderr)

    def tag(self, source_image: Union[Image, str], new_tag: str):
        """Adds a tag to a Docker image.

        Alias: `docker.tag(...)`

        # Arguments
            source_image: The Docker image to tag. You can use a tag to reference it.
            new_tag: The tag to add to the Docker image.

        # Raises
            `dockertown.exceptions.NoSuchImage` if the image does not exists.
        """
        full_cmd = self.docker_cmd + [
            "image",
            "tag",
            str(source_image),
            new_tag,
        ]
        run(full_cmd)

    def _pull_if_necessary(self, image: ValidImage) -> Image:
        if isinstance(image, Image):
            # A docker image object must exist in the daemon to be defined.
            return image
        try:
            return self.inspect(image)
        except DockerException:
            print(f"Unable to find image '{image}' locally")
            return self.pull(image)

    def copy_from(
        self,
        image: ValidImage,
        path_in_image: ValidPath,
        destination: ValidPath,
        pull: str = "missing",
    ):
        with container_cli_wrapper.ContainerCLI(self.client_config).create(
            image, pull=pull
        ) as tmp_container:
            tmp_container.copy_from(path_in_image, destination)

    def copy_to(
        self,
        base_image: ValidImage,
        local_path: ValidPath,
        path_in_image: ValidPath,
        new_tag: Optional[str] = None,
        pull: str = "missing",
    ) -> Image:
        with container_cli_wrapper.ContainerCLI(self.client_config).create(
            base_image, pull=pull
        ) as tmp_container:
            tmp_container.copy_to(local_path, path_in_image)
            return tmp_container.commit(tag=new_tag)
